/*****************************************************
 **
 **  Remote Memory Library - SERVER
 **
 **  File: MemoryServer.cpp
 **  Author: Marcio S. Pinho e Rafael S. Garcia.
 **  Mail: pinho@inf.pucrs.br,
 **        rafael_sgarcia@terra.com.br
 **  Version: 2.0.
 **
 **  Last Update: 27.January.2003
 **
 *****************************************************/
#include <stdio.h>
#include <conio.h>
#include <winsock.h>
#include "RemoteMemory.h"

SOCKADDR_IN clientSockAddr;
SOCKET clientSocketList[100];
int clientNumber = 0;

char Memory[1024];





/************************************************
 ** IsThereAMsg ( SOCKET )
 ************************************************/
BOOL IsThereAMsg ( SOCKET sock ) {
	fd_set socketSet;
	const struct timeval FAR timeout = { 0, 0 };
	int result;

	FD_ZERO ( &socketSet );
	FD_SET ( sock, &socketSet );

	result = select ( 1, &socketSet, NULL, NULL, &timeout );
	if ( FD_ISSET ( sock, &socketSet ) )
		return true;

	return false;
}



/************************************************
 ** WriteOnLocalMemory ( *void, WORD, int )
 ************************************************/
void WriteOnLocalMemory ( void *msg, WORD addr, int n ) {
	memcpy ( &Memory [addr], msg, n );
}



/************************************************
 ** ReadFromLocalMemory ( *void, WORD, int )
 ************************************************/
void ReadFromLocalMemory ( void *msg, WORD addr, int n ) {
	memcpy ( msg, &Memory [addr], n );
}





/************************************************
 ** Remote Memory IP Server Functions
 ************************************************/


/************************************************
 ** CreateServerIP ( unsigned short )
 ************************************************/
SOCKET CreateServerIP ( unsigned short portNum ) { 
	char myName[265];
	HOSTENT *hostent;
	WSAData info;

	SOCKET serverSocket;
	SOCKADDR_IN socketAddr;

	WORD socketVersion = MAKEWORD ( 1, 1 );

	// Setup the socket library version.
	if ( WSAStartup ( socketVersion, &info ) == SOCKET_ERROR ) {
		printf ( "Couldn't initialize the socket library.\n" );
		exit (1);
	}

	// Check the library version.
	if ( info.wVersion != socketVersion ) {
		fprintf ( stderr, "Couldn't initialize the socket library.\n" );
		exit (1);
	}

	// clear the address struct.
	memset ( &socketAddr, 0, sizeof ( struct sockaddr_in ) );

	int result = gethostname ( myName, sizeof ( myName ) );
	if ( result == WSANOTINITIALISED )
		printf ( "Socket library wasn't correctly initialized.\n" );

	// Check our own address.
	hostent = gethostbyname ( myName );
	if ( hostent == NULL )
		return INVALID_SOCKET;

	// Create the socket.
	serverSocket = socket ( AF_INET, SOCK_STREAM, 0 );
	if ( serverSocket == INVALID_SOCKET )
		return INVALID_SOCKET;

	socketAddr.sin_family = hostent->h_addrtype;
	socketAddr.sin_port = htons ( portNum );

	// Bind the socket.
	if ( bind ( serverSocket, ( struct sockaddr * )&socketAddr, sizeof ( struct sockaddr_in ) ) == SOCKET_ERROR ) {
		closesocket ( serverSocket );
		return INVALID_SOCKET;
	}

	listen ( serverSocket, SOMAXCONN );
	printf ( "TCP/IP Server created at port %d.\n", portNum );
	return serverSocket;

} 



/************************************************
 ** SendMessageIP ( SOCKET, *char, int )
 ************************************************/
void SendMessageIP ( SOCKET sock, char *buf, int numBytes ) {
	int result = send ( sock, buf, numBytes, 0 );
}



/************************************************
 ** ReceiveMessageIP ( SOCKET, *char, int )
 ************************************************/
void ReceiveMessageIP ( SOCKET sock, char *buf, int numBytes ) {
	memset ( buf, 0, numBytes );
	int result = recv ( sock, buf, numBytes, 0 );
	if ( result == SOCKET_ERROR ) {
		printf ( "Error reading socket number %d\n", sock );
		exit (1);
	}
}



/************************************************
 ** CloseAllClientsIP ( void )
 ************************************************/
void CloseAllClientsIP ( void ) {
	for ( int counter = 0; counter < clientNumber; counter++ )
		closesocket ( clientSocketList[counter] );
}



/************************************************
 ** AcceptNewClientIP ( SOCKET )
 ************************************************/
void AcceptNewClientIP ( SOCKET serverSocket ) {
	SOCKET newSocket = accept ( serverSocket, NULL, NULL );
	if ( newSocket == INVALID_SOCKET ) {
		fprintf ( stderr, "Error waiting for new connection!\n" );
		closesocket ( serverSocket );
		exit (1);
	}
	
	printf ( "New client received !\n" );
	clientSocketList[clientNumber] = newSocket;
	clientNumber++;
}



/************************************************
 ** ReceiveMessageToRemoteMemoryIP ( SOCKET )
 ************************************************/
void ReceiveMessageToRemoteMemoryIP ( SOCKET sock ) {
	TMSG Package;

	ReceiveMessageIP ( sock, ( char * )&Package, sizeof ( TMSG ) );
	if ( Package.nbytes > MAX_BUFF_SIZE ) {
		printf ( "Error in 'ReceiveMessageToRemoteMemory': The maximum message size is %d\n", MAX_BUFF_SIZE );
		exit (1);
	}

    if ( Package.nbytes == 0 ) {
		printf ( "Error reading socket number %d\n", sock );
		exit (1);
	}

	switch (Package.operation) {
		case REMOTE_MEMORY_WRITE:				
			memcpy ( &Memory[Package.address], Package.data, Package.nbytes );
			break;

		case REMOTE_MEMORY_READ:
			Package.operation = REMOTE_MEMORY_REPLY;
			memcpy ( Package.data, &Memory[Package.address], Package.nbytes );
			SendMessageIP ( sock, ( char * )&Package, sizeof ( TMSG ) );		
			break;
	}
}



/************************************************
 ** ReceiveRequestFromClientsIP ( void )
 ************************************************/
void ReceiveRequestsFromClientsIP ( void ) {
	Sleep (1);

	for ( int counter = 0; counter < clientNumber; counter++ ) {
		if ( IsThereAMsg ( clientSocketList[counter] ) )
			ReceiveMessageToRemoteMemoryIP ( clientSocketList[counter] );
	}
}





/************************************************
 ** Remote Memory UDP Server Functions
 ************************************************/

/************************************************
 ** CreateServerUDP ( unsigned short )
 ************************************************/
SOCKET CreateServerUDP ( unsigned short portNum ) {
	char myName[265];
	HOSTENT *hostent;
	WSAData info;

	SOCKET serverSocket;
	SOCKADDR_IN socketAddr;

	WORD socketVersion = MAKEWORD ( 1, 1 );

	// Setup the socket library version.
	if ( WSAStartup ( socketVersion, &info ) == SOCKET_ERROR ) {
		printf ( "Couldn't initialize the socket library.\n" );
		exit (0);
	}

	// Check the library version.
	if ( info.wVersion != socketVersion ) {
		fprintf ( stderr, "Couldn't initialize the socket library.\n" );
		exit (0);
	}

	// clear the address struct.
	memset ( &socketAddr, 0, sizeof ( struct sockaddr_in ) );

	int result = gethostname ( myName, sizeof ( myName ) );
	if ( result == WSANOTINITIALISED )
		printf ( "Socket library wasn't correctly initialized.\n" );

	// Check our own address.
	hostent = gethostbyname ( myName );
	if ( hostent == NULL )
		return INVALID_SOCKET;

	// Create the socket.
	serverSocket = socket ( AF_INET, SOCK_DGRAM, 0 );
	if ( serverSocket == INVALID_SOCKET )
		return INVALID_SOCKET;

	socketAddr.sin_family = hostent->h_addrtype;
	socketAddr.sin_port = htons ( portNum );
	socketAddr.sin_addr.s_addr = htonl ( INADDR_ANY );

	// Bind the socket.
	if ( bind ( serverSocket, (struct sockaddr *)&socketAddr, sizeof ( struct sockaddr_in ) ) == SOCKET_ERROR ) {
		closesocket ( serverSocket );
		return INVALID_SOCKET;
	}

	printf ( "TCP/UDP Server created at port %d.\n", portNum );
	return serverSocket;
}



/************************************************
 ** SendMessageUDP ( SOCKET, *char, int )
 ************************************************/
void SendMessageUDP ( SOCKET sock, char *buf, int numBytes ) {
	int result = sendto ( 
		sock, buf, numBytes, 0, ( struct sockaddr *)&clientSockAddr, sizeof ( clientSockAddr ) 
	);
}



/************************************************
 ** ReceiveMessageUDP ( SOCKET, *char, int )
 ************************************************/
void ReceiveMessageUDP ( SOCKET sock, char *buf, int numBytes ) {
	memset ( buf, 0, numBytes );
	int length = sizeof ( clientSockAddr );
	int result = recvfrom ( sock, buf, numBytes, 0, ( struct sockaddr *)&clientSockAddr, &length );
	if ( result == SOCKET_ERROR ) {
		printf ( "Error reading socket number %d\n", sock );
		exit (0);
	}
}



/************************************************
 ** ReceiveMessageToRemoteMemoryUDP ( SOCKET )
 ************************************************/
void ReceiveMessageToRemoteMemoryUDP ( SOCKET sock ) {
	TMSG Package;

	ReceiveMessageUDP ( sock, ( char * )&Package, sizeof ( TMSG ) );
	if ( Package.nbytes > MAX_BUFF_SIZE ) {
		printf ( "Error in 'ReceiveMessageToRemoteMemory': The maximum message size is %d;\n", MAX_BUFF_SIZE );
		exit (1);
	}

	if ( Package.nbytes == 0 ) {
		printf ( "Error reading socket number %d.\n", sock );
		exit (1);
	}

	switch ( Package.operation ) {
		case REMOTE_MEMORY_WRITE:
			memcpy ( &Memory[Package.address], Package.data, Package.nbytes );
			break;

		case REMOTE_MEMORY_READ:
			Package.operation = REMOTE_MEMORY_REPLY;
			memcpy ( Package.data, &Memory[Package.address], Package.nbytes );
			SendMessageUDP ( sock, (char *)&Package, sizeof (TMSG) );
			break;
	}
}



/************************************************
 ** main ( int, *char )
 ************************************************/
int main ( int argc, char *argv[] ) {
	SOCKET serverSocketIP;
	SOCKET serverSocketUDP;
	int key;
	
	if ( argc < 2 ) {
		printf ("Command line parameters: port.\n" );
		printf ("Press any key to exit\n");
		getch();
		exit (0);
	}

	int port = atoi ( argv[1] );
	printf("************************************************\n");
	printf("            MemorySocket Server\n", port);
	printf("************************************************\n");

	serverSocketIP  = CreateServerIP  ( port );
	serverSocketUDP = CreateServerUDP ( port + 1 );
	printf ( "Press ESC key to exit\n" );

	while ( true ) {
		if ( kbhit() ) {
			key = getch();
			if ( key == ESC ) {
				closesocket ( serverSocketIP );
				closesocket ( serverSocketUDP );
				exit (0);
			}
		}

		// Verify messages to the IP server.
		if ( IsThereAMsg ( serverSocketIP ) )
			AcceptNewClientIP ( serverSocketIP );

		// Verify messages to the UDP server.
		if ( IsThereAMsg ( serverSocketUDP ) )
			ReceiveMessageToRemoteMemoryUDP ( serverSocketUDP );

		// Check request of the IP clients.
		ReceiveRequestsFromClientsIP();
	}
	
	closesocket ( serverSocketIP );
	closesocket ( serverSocketUDP );
	CloseAllClientsIP();
	return 0;
}

